def confirm(question: str) -> bool:
  """confirm in cli process

  Args:
      question (str): information of question

  Returns:
      bool: True to comfirm
  """
  answer = ''
  while answer not in ['y', 'yes', 'n', 'no']:
    answer = input(f'{question}[Y/N]').lower()
  return answer in ['y', 'yes']


from typing import Any, Callable, Dict, List, Tuple, Union
import logging

from research_utils.sqlite.utils import dict_to_values_sql
from ..decorators.logging import logger
from pathlib import Path
import sqlite3
from sqlite3 import Error, Connection, Cursor
from .typing.sql import FieldOption
from .typing.operator import ConditionSQLArgument, OperatorSQLArgument


def create_db_if_not_exist(database) -> None:
  file = Path(database)
  file.parents[0].mkdir(parents=True, exist_ok=True)
  file.touch(exist_ok=True)
  logging.debug(f'{file} is created')


@logger()
def create_connection(database=':memory:', *args, **kwargs) -> Connection:
  """Create a Database Connection
  Args:
      database: {StrOrBytesPath} -- use ":memory:" to open a database 
          connection to a database that resides in RAM instead of on disk
      check_same_thread: {bool} -- If set False, the returned connection 
          may be shared across multiple threads {default: {True}}
  Returns:
      Connection: ...
  """

  if database != ':memory:':
    create_db_if_not_exist(database)

  conn = None
  try:
    conn = sqlite3.connect(database, *args, **kwargs)
  except Error as e:
    logging.error(e)
  return conn


def create_table(conn: Connection, table_name: str, fields: Dict[str, FieldOption]):
  """Create Table

  Args:
      conn (Connection): The Connection object
      table_name (str): Name of table
      fields (Dict[str, FieldOption]): Dictionary describing information of the fields
  """
  sql = ''.join((f'CREATE TABLE IF NOT EXISTS {table_name} (',
                 *(f'{key} {value.sql}{"," if index != len(fields) - 1 else ""}\n'
                   for index, (key, value) in enumerate(fields.items())), ');'))
  logging.info(sql)
  try:
    cur = conn.cursor()
    cur.execute(sql)
  except Error as e:
    logging.error(e)


def delete_table(conn: Connection, table_name: str):
  sql = f'DROP TABLE {table_name}'
  logging.info(sql)
  try:
    cur = conn.cursor()
    cur.execute(sql)
    conn.commit()
  except Error as e:
    logging.error(e)


def clear_table(conn: Connection, table_name: str):
  sql = f'TRUNCATE TABLE {table_name}'
  logging.info(sql)
  try:
    cur = conn.cursor()
    cur.execute(sql)
    conn.commit()
  except Error as e:
    logging.error(e)


def insert(conn: Connection, table_name: str, fields: Union[List[str], Tuple[str]],
           values: Union[List[Tuple], Tuple]):
  """Create Row or Rows into Custom Table

  Args:
      conn (Connection): The Connection object
      table_name (str): Name of table
      fields (Union[List[str], Tuple[str]]): 
      values (Union[List[Tuple], Tuple]): 
  """
  sql = f"""INSERT INTO {table_name} ({', '.join(map(str, fields))})
  VALUES
  ({', '.join(['?' for _ in fields])})
  """
  logging.info(sql)
  try:
    cur = conn.cursor()
    if isinstance(values, tuple):
      cur.execute(sql, values)
    else:
      cur.executemany(sql, values)
    conn.commit()
  except Error as e:
    logging.error(e)


@logger()
def query(conn: Connection,
          table_name: str,
          fields: Union[List[str], Tuple[str]] = None,
          row_factory: Callable[[Cursor, Tuple], Any] = None,
          where: Union[OperatorSQLArgument, ConditionSQLArgument] = None) -> List:
  """Query All Rows from Custom Table

  Args:
      conn (Connection): The Connection object
      table_name (str): Name of table
      fields (Union[List[str], Tuple[str]], optional): Defaults to None.
      where (Union[OperatorSQLArgument, ConditionSQLArgument], optional): Defaults to None.

  Returns:
      List: List of rows
  """

  sql = f"""SELECT {', '.join(map(str, fields)) if fields is not None and len(fields) > 0 else '*'} FROM {table_name}
  {f'WHERE {where.sql}' if where is not None else ''}
  """
  logging.info(sql)

  if row_factory is not None:
    conn.row_factory = row_factory

  try:
    cur = conn.cursor()
    cur.execute(sql)
    rows = cur.fetchall()
    return rows
  except Error as e:
    logging.error(e)


@logger()
def update(conn: Connection,
           table_name: str,
           data: Dict[str, Any],
           where: Union[OperatorSQLArgument, ConditionSQLArgument] = None):
  """Update rows in custom table

  Args:
      conn (Connection): The Connection object
      table_name (str): Name of table
      data (Dict[str, Any]): dict of fields and values
      where (Union[OperatorSQLArgument, ConditionSQLArgument], optional): Defaults to None.
  """

  if not data or not len(data):
    # do nothing if data is None or no values to be updated
    print(data)
    pass
  sql = f"""UPDATE {table_name}
  SET {dict_to_values_sql(data)}
  {f'WHERE {where.sql}' if where is not None else ''}
  """
  logging.info(sql)

  try:
    cur = conn.cursor()
    cur.execute(sql)
    conn.commit()
  except Error as e:
    logging.error(e)


import logging
from typing import Dict, List, Union
from numbers import Number
from itertools import chain
from tempfile import TemporaryFile
from pathlib import Path
import yaml
import numpy as np
from research_utils.decorators.logging import logger


def str_to_class(module_name: str, class_name: str):
  import importlib
  module = importlib.import_module(module_name)
  return getattr(module, class_name)


@logger()
def load_yaml(path: Union[str, Path]) -> Dict:
  with open(path, "r") as stream:
    try:
      return yaml.safe_load(stream)
    except yaml.YAMLError as e:
      logging.error(e)


def flatten(raw: List[Union[tuple, Number]]):
  # [reference](https://stackoverflow.com/a/22569169)
  return list(chain(*(i if isinstance(i, tuple) else (i, ) for i in raw)))


def npy2tmpfile(data: np.ndarray):
  # [reference](https://numpy.org/doc/stable/reference/generated/numpy.save.html)
  file = TemporaryFile()
  np.save(file, data)
  _ = file.seek(0)
  return file


def conditional_kwargs(*args, conditional: bool = True, **kwargs):
  """return custom kwargs while conditional is True

  Args:
      conditional (bool): if conditional is True, return kwargs

  Returns:
      dict: kwargs
  """
  if conditional:
    return kwargs
  return {}


def dict_union(*args):
  return dict(chain.from_iterable(d.items() for d in args))


def ensure_path_existed(*args) -> Path:
  path = Path(*args)
  path.mkdir(parents=True, exist_ok=True)
  return path



from typing import Callable, List, Dict
from torch.utils.data.dataset import Dataset
from research_utils.sqlite.helper import Helper, FieldType
from research_utils.sqlite.typing.sql import SQLArgument


class DBFolder(Dataset):
  def __init__(self, db_path: str, table_name: str, fields: Dict[str, FieldType],
               argument: SQLArgument) -> None:
    super().__init__()
    helper = Helper(db_path=db_path, table_name=table_name, fields=fields)
    self.__rows = helper.query(fields=list(fields.keys()), argument=argument)

  def __len__(self):
    return len(self.__rows)

  def __getitem__(self, index):
    return self.__rows[index]


from typing import Union
import os
from pathlib import Path
from mlflow.tracking.client import MlflowClient
from research_utils.typing import PathLike
from research_utils.decorators.logging import logger, Logging_Level


@logger(level=Logging_Level.INFO, transform=lambda path: f'Artifacts is downloaded at {path}')
def download_artifacts(run_id: str,
                       remote_path: Union[str, Path],
                       local_path: Union[str, Path],
                       client: MlflowClient = MlflowClient(),
                       no_cached=False) -> PathLike:
  if os.path.exists(local_path) and no_cached:
    os.remove(Path(local_path, remote_path))
  os.makedirs(local_path, exist_ok=True)
  return client.download_artifacts(run_id=run_id, path=remote_path, dst_path=local_path)


def convertToBinaryData(filename):
  with open(filename, 'rb') as file:
    blob = file.read()
  return blob


def extractDictByKeys(dict: dict, fields):
  return [dict.get(field) for field in fields]


def dict_to_values_sql(data: dict):
  return ', '.join(map(lambda item: f'{item[0]}="{str(item[1])}"', data.items()))


import sys
from typing import Any, Callable, Dict
import functools
from enum import Enum
import logging

grey = "\x1b[38;21m"
green = "\x1b[1;32m"
yellow = "\x1b[33;21m"
red = "\x1b[31;21m"
bold_red = "\x1b[31;1m"
reset = "\x1b[0m"

class Logging_Level(Enum):
  DEBUG = logging.DEBUG
  INFO = logging.INFO
  WARNING = logging.WARNING
  ERROR = logging.ERROR
  CRITICAL = logging.CRITICAL

class CustomFormatter(logging.Formatter):
  fmt = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

  def __init__(self, fmt: str = fmt, datefmt: str = None, style = '%') -> None:
      super().__init__(fmt=fmt, datefmt=datefmt, style=style)
      self.FORMATS = {
          logging.DEBUG: grey + fmt + reset,
          logging.INFO: green + fmt + reset,
          logging.WARNING: yellow + fmt + reset,
          logging.ERROR: red + fmt + reset,
          logging.CRITICAL: bold_red + fmt + reset
      }

  def format(self, record: logging.LogRecord):
      log_fmt = self.FORMATS.get(record.levelno)
      formatter = logging.Formatter(log_fmt)
      return formatter.format(record)

def log(*args, level: int, item, **kwargs):
  level = Logging_Level(level)
  if level == Logging_Level.DEBUG:
    logging.debug(item)
  elif level == Logging_Level.INFO:
    logging.info(item)
  elif level == Logging_Level.WARNING:
    logging.warning(item)
  elif level == Logging_Level.ERROR:
    logging.error(item)
  elif level == Logging_Level.CRITICAL:
    logging.critical(item)
  else:
    pass

def config_logger(level: Logging_Level=Logging_Level.WARNING, format='[%(asctime)s %(levelname)-8s] %(message)s', datefmt='%Y%m%d %H:%M:%S'):
  def decorator(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
      logger = logging.getLogger()
      logger.setLevel(level.value)
      if len(logger.handlers) == 0:
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(level.value)
        handler.setFormatter(CustomFormatter())
        logger.addHandler(handler)
      return func(*args, **kwargs)
    return inner
  return decorator

def logger(
  log_func: Callable[[int, Any], None]=None, 
  level: Logging_Level=Logging_Level.DEBUG, 
  transform: Callable=None
):
  log_func = log if log_func is None else log_func
  def decorator(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
      items = func(*args, **kwargs)
      try:
        for item in items.items() if isinstance(items, dict) else items:
          log_func(*args, level=level.value, item=item if transform is None else transform(item), **kwargs)
      except:
        log_func(*args, level=level.value, item=items if transform is None else transform(items), **kwargs)
      return items
    return inner
  return decorator

def conditional_logger(
  conditional: Callable[[Any, Dict[str, Any]], bool] = lambda **kwargs: True, 
  log_func: Callable[[tuple, Dict[str, Any]], None]=log, 
  level: Logging_Level=Logging_Level.DEBUG, 
  transform: Callable=None
):
  """conditional logger

  Args:
      conditional (Callable[[Any, Dict[str, Any]], bool], optional): only kwargs will be passed into. Defaults to lambda**kwargs:True.
      log_func (Callable[[tuple, Dict[str, Any]], None], optional): Defaults to log.
      level (Logging_Level, optional): Defaults to Logging_Level.DEBUG.
      transform (Callable, optional): Defaults to None.

  Returns:
      (func) -> (*args, **kwargs) -> Any: return a decorator
  """
  log_func = log if log_func is None else log_func
  def decorator(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
      def _log_func(item: Any):
        log_func(*args, level=level.value, item=item if transform is None else transform(item), **kwargs)
      items = func(*args, **kwargs)
      if conditional(items, **kwargs):
        try:
          for item in items.items() if isinstance(items, dict) else items:
            _log_func(item)
        except:
          _log_func(items)
      return items
    return inner
  return decorator




from typing import Any, Sequence, Union
from argparse import Action, ArgumentParser, Namespace
from enum import Enum

class EnumAction(Action):
  """Argparser Action for Handling Enums
  References:
    [Support for Enum arguments in argparse](https://stackoverflow.com/a/60750535)
  """
  def __init__(self, **kwargs) -> None:
    # Pop off the type value
    enum_type = kwargs.pop('type', None)

    # Ensure an Enum subclass is provided
    if enum_type is None:
      raise ValueError('Type must be assigned an Enum when using EnumAction')
    if not issubclass(enum_type, Enum):
      raise TypeError('Type must be an Enum when using EnumAction')

    # Generate choices from the Enum
    kwargs.setdefault("choices", tuple(e.name for e in enum_type))
    super().__init__(**kwargs)

    self._enum = enum_type

  def __call__(self, parser: ArgumentParser, namespace: Namespace, values: Union[str, Sequence[Any]], option_string: str = None) -> None:
    # Convert value back into an Enum
    value = self._enum[values]
    setattr(namespace, self.dest, value)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['research_utils',
 'research_utils.argparse',
 'research_utils.argparse.actions',
 'research_utils.decorators',
 'research_utils.mlflow',
 'research_utils.sqlite',
 'research_utils.sqlite.typing',
 'research_utils.torch',
 'research_utils.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'numpy>=1.21.5,<2.0.0', 'sqlite-utils>=3.6,<4.0']

extras_require = \
{'mlflow': ['mlflow>=1.22.0,<2.0.0'], 'torch': ['torch>=1.10.1,<2.0.0']}

setup_kwargs = {
    'name': 'research-utils',
    'version': '0.0.5',
    'description': 'some utils for my research',
    'long_description': '# research-utils\n[![Build](https://github.com/r08521610/research-utils/actions/workflows/package-build.yml/badge.svg)](https://github.com/r08521610/research-utils/actions/workflows/package-build.yml)\n[![Publish Package](https://github.com/r08521610/research-utils/actions/workflows/package-publish.yml/badge.svg?branch=main)](https://github.com/r08521610/research-utils/actions/workflows/package-publish.yml)\n[![PyPI version](https://badge.fury.io/py/research-utils.svg)](https://badge.fury.io/py/research-utils)\n',
    'author': 'Rainforest Cheng',
    'author_email': 'r08521610@ntu.edu.tw',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://r08521610.github.io/research-utils/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<3.10',
}


setup(**setup_kwargs)
